package choco;

import org.chocosolver.solver.Model;
import org.chocosolver.solver.Solver;
import org.chocosolver.solver.constraints.nary.cumulative.Cumulative;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.solver.variables.Task;

import static org.chocosolver.solver.search.strategy.Search.minDomLBSearch;

public class job_shop {

	public static void main(String[] args) {				
/*
	//jobshop data file: simple.txt 
		int njobs = 4; int ntsks = 3; int maxDuration = 7;
		int [][] durs = {{2,4,7},{3,4,5},{5,3,3},{3,4,4}};
		int [][] mchs = {{0,1,2},{2,1,0},{0,1,2},{1,2,0}};
*/
	// jobshop data file: js_ft06.txt
		int njobs = 6; int ntsks = 6;
		int nmchs = ntsks;
		int [][] durs = {{1,3,6,7,3,6},{8,5,10,10,10,4}, {5,5,8,9,1,7}, {5,5,5,3,8,9},{9,3,5,4,3,1}, {3,3,9,10,4,1}};
		int [][] mchs = {{2,0,1,3,5,4},{1,2,4,5,0,3}, {2,3,5,0,1,4},{1,0,2,3,4,5},{2,1,4,5,0,3}, {1,3,5,0,4,2}};
		
		int maxDur = durs[0][0]; // obtain maximum duration of any task
		for(int i=0; i<njobs; i++) for (int j=0; j<ntsks; j++) if (durs[i][j]>maxDur) maxDur = durs[i][j];
		int maxSpan = njobs*ntsks* maxDur;
		
	// search options
		// stop search
			//int backLimit = 1000; s.limitBacktrack(1000);
			//int solLimit = 15; s.limitSolution(solLimit);
			//int timeLimit = 120000; s.limitTime(timeLimit);
		
	/*	if cumulative constraint is used (see below):
			graphBased = true | false
			cum:filtering algorithm (see below)
				cum = Cumulative.Filter.C where
				C = DEFAULT | DISJUNCTIVE_TASK_INTERVAL | HEIGHTS | NAIVETIME | NRJ | SWEEP | SWEEP_HEI_SORT | TIME;			 
	*/
		boolean cumul = true;
		boolean graphBased = false;  	
		Cumulative.Filter cum = Cumulative.Filter.TIME;
		
	// Whether first fail heuristics (minDomLBSearch)is used 
		boolean ff = false; 
		
	// Now the model		
		Model md = new Model("(" + njobs +"," + nmchs + ") - Job Shop");
		Solver sv = md.getSolver();
		
	// decision variables declared in an array to help heuristics
		// starts[i*ntsks+j] is the start of task j of job i 
		IntVar [] starts = md.intVarArray("N", njobs*ntsks, 1, maxSpan);
				
	// tasks indexed by job and task
		Task [][] tx = new Task[njobs][ntsks];
		for(int i = 0; i < njobs; i++){
			for (int j = 0; j < ntsks; j++){
				tx[i][j] = new Task(starts[i*ntsks+j], md.intVar(durs[i][j],durs[i][j]), md.intVar(1, maxSpan));}}

	// tasks indexed by machine (and task-job)
		Task [][] mx = new Task[nmchs][njobs];
		for (int k = 0; k < nmchs; k++){
			int p = 0;
			for(int i = 0; i < njobs; i++){
				for (int j = 0; j < ntsks; j++){
					if(mchs[i][j] == k){
						mx[k][p] = tx[i][j] ;
						p = p + 1; }}}}

	// precedence constraints
		for (int i = 0; i < njobs; i++) {
			for (int j = 0; j < ntsks-1; j++) {
				md.arithm(tx[i][j].getEnd(), "<=", tx[i][j+1].getStart()).post(); }}
		
	// obtain the end of last task
		IntVar [] lx = md.intVarArray(njobs,1,maxSpan);
		IntVar last = md.intVar(1,maxSpan);
		for (int i = 0; i < njobs; i++)
			lx[i] = tx[i][ntsks-1].getEnd();
		md.max(last, lx).post();
		
	// Decision Problem: makeSpan is acceptable - less than some constant acceptSpan
		// int acceptSpan = ...;
		// md.arithm(last, "<=", acceptSpan).post();
	// Optimization Problem: set objective as minimize last		
		md.setObjective(Model.MINIMIZE, last);	
	// Uses heuristics or not
		if (ff) sv.setSearch(minDomLBSearch(starts));

		if (cumul){
			IntVar one = md.intVar(1,1);
			IntVar [] ones = md.intVarArray(njobs,1,1);
			for (int k = 0; k < nmchs; k++)
				md.cumulative(mx[k], ones, one, graphBased, cum).post();
		} else {
			for (int k = 0; k < nmchs; k++) {
				for (int i = 0; i < mx[k].length-1; i++) {
					for (int j = i+1; j < mx[k].length; j++) {
						md.or(
							md.arithm(mx[k][i].getEnd(), "<=", mx[k][j].getStart()),
							md.arithm(mx[k][j].getEnd(), "<=", mx[k][i].getStart())
						).post();}}}}		
		
		while (sv.solve()){
			System.out.print("\n\nSolving " + sv.getModelName() + " (solution " + sv.getSolutionCount() + ");");
			System.out.println(" with Makespan = " + last.getValue() + " after " + sv.getTimeCount() + " secs");

			System.out.println("--- Jobs ---");
			for(int i = 0; i< njobs; i++){	
				for (int j = 0; j< ntsks; j++) {
					System.out.print(" " + tx[i][j].getStart().getValue() + " to " + tx[i][j].getEnd().getValue()+"; ");}
				System.out.println();}
			System.out.println();
			System.out.println("--- Machines ---");
			for(int i = 0; i< nmchs; i++){	
				for (int j = 0; j< ntsks; j++) {
					System.out.print(" " + mx[i][j].getStart().getValue() + " to " + mx[i][j].getEnd().getValue()+"; ");}
				System.out.println();}
		//sv.printStatistics();
		}
		System.out.println("\n !!! No (more) solutions after " + sv.getTimeCount() + " secs");}
}	

/*	for each type of machine:
		cumulative(tasks, machines/task, total_machines, graphBased, cum)
where:
	machines/task = ones, since each task uses one machine
	total_machines = one, since there is only one machine of each type
	graphBased = true/false (see below) 
	cum: type of filtering algorithm used (see below)  
		DEFAULT / DISJUNCTIVE_TASK_INTERVAL / HEIGHTS / NAIVETIME / NRJ / SWEEP / SWEEP_HEI_SORT / TIME
filters:
• DEFAULT: Combines above filters as a black-box not idempotent
• DISJUNCTIVE_TASK_INTERVAL: energetic reasoning to filter disjunctive constraint Only propagated on variable subsets of size < 30 not idempotent not enough to ensure correctness (only an additional filtering)
• HEIGHTS:filters height variables only (sweep-based algorithm) idempotent (on the given set of variables only)
• NAIVETIME:time-table algorithm based on each point in time not idempotent
• NRJ: energetic reasoning to filter not idempotent not enough to ensure correctness (only an additional filtering)
• SWEEP: time-table algorithm based on a sweep line idempotent (on the given set of variables only)
• SWEEP_HEI_SORT: time-table algorithm based on a sweep line idempotent (on the given set of variables only)
• TIME: time-table algorithm based on each point in time not idempotent		

graphBased - parameter indicating how to filter: 
 • TRUE: applies on subset of overlapping tasks
 • FALSE: applies on all tasks
*/

