package choco;

import static org.chocosolver.solver.search.strategy.Search.minDomLBSearch;

import java.io.IOException;
import org.chocosolver.solver.Model;
import org.chocosolver.solver.Solver;
import org.chocosolver.solver.constraints.nary.cumulative.Cumulative;
import org.chocosolver.solver.variables.IntVar;
import org.chocosolver.solver.variables.Task;


public class job_shop_file {

	public static void main(String[] args)  throws IOException {
		jobs js = new jobs();
		//js.ReadJobs("src/choco/jobshop/js_orb01.txt");
		//js.ReadJobs("src/choco/jobshop/js_abz07.txt");
		//js.ReadJobs("src/choco/jobshop/js_la01.txt");
		js.ReadJobs("src/choco/jobshop/js_ft06.txt");
				
		int njobs = js.nj;
		int ntsks = js.nt;
		int nmchs = ntsks;
		int [][] tasks = js.tasks;
		int [][] machs = js.machs;
		
		
		int maxDur = tasks[0][0]; // obtain maximum duration of any task
		for(int i=0; i<njobs; i++) for (int j=0; j<ntsks; j++) if (tasks[i][j]>maxDur) maxDur = tasks[i][j];
		int maxSpan = njobs*ntsks* maxDur;	
		
	// search options		
		//int backLimit = 1000; s.limitBacktrack(1000);
		//int solLimit = 15; s.limitSolution(solLimit);
		//int timeLimit = 120000; s.limitTime(timeLimit);
		boolean cumul = true;
		boolean graphBased = true;
		Cumulative.Filter cum = Cumulative.Filter.DISJUNCTIVE_TASK_INTERVAL;
		
		//System.out.println("njobs = " + Integer.toString(njobs));
		//System.out.println("ntsks = " + Integer.toString(ntsks));
		// Whether first fail heuristics (minDomLBSearch)is used 
		boolean ff = false; 
		
	// Now the model			
		Model md = new Model("(" + njobs +"," + nmchs + ") - Job Shop");
		Solver sv = md.getSolver();

		
		IntVar [] starts = md.intVarArray("N", njobs*ntsks, 1, maxSpan);
		
		// tasks indexed by job and task
		Task [][] tx = new Task[njobs][ntsks];
		for(int i = 0; i < njobs; i++){
			for (int j = 0; j < ntsks; j++){
				//System.out.println(i*njobs+j);
				tx[i][j] = new Task(starts[i*ntsks+j], md.intVar(tasks[i][j],tasks[i][j]), md.intVar(1, maxSpan));
			}
		}
		
		// tasks indexed by machine and job?
		Task [][] mx = new Task[nmchs][njobs];
		for (int k = 0; k < nmchs; k++){
			int p = 0;
			for(int i = 0; i < njobs; i++){
				for (int j = 0; j < ntsks; j++){
					if(machs[i][j] == k){
						mx[k][p] = tx[i][j];
						p = p + 1;
					}
				}
			}			
		}

		// precedence constraints
		for (int i = 0; i < njobs; i++) {
			for (int j = 0; j < ntsks-1; j++) {
				md.arithm(tx[i][j].getEnd(), "<=", tx[i][j+1].getStart()).post();
			}
		}

		// set objective 
		IntVar [] lx = md.intVarArray(njobs,1,maxSpan);
		IntVar last = md.intVar(1,maxSpan);
		for (int i = 0; i < njobs; i++)
			lx[i] = tx[i][ntsks-1].getEnd();
		md.max(last, lx).post();
		md.setObjective(Model.MINIMIZE, last);
		if (ff)
			sv.setSearch(minDomLBSearch(starts));

		if (cumul){
			IntVar one = md.intVar(1,1);
			IntVar [] ones = md.intVarArray(njobs,1,1);
			for (int k = 0; k < nmchs; k++)
				md.cumulative(mx[k], ones, one, graphBased, cum).post();
		} else {
			for (int k = 0; k < nmchs; k++) {
				for (int i = 0; i < mx[k].length-1; i++) {
					for (int j = i+1; j < mx[k].length; j++) {
						md.or(
							md.arithm(mx[k][i].getEnd(), "<=", mx[k][j].getStart()),
							md.arithm(mx[k][j].getEnd(), "<=", mx[k][i].getStart())
						).post();
					}
				}
			}
		}
		
		while (sv.solve()){
			System.out.print("\n\nSolving " + sv.getModelName() + " (solution " + sv.getSolutionCount() + ");");
			System.out.println(" with Makespan = " + last.getValue() + " after " + sv.getTimeCount() + " secs");

			System.out.println("--- Jobs ---");
			for(int i = 0; i< njobs; i++){	
				for (int j = 0; j< ntsks; j++) {
					System.out.print(" " + tx[i][j].getStart().getValue() + " to " + tx[i][j].getEnd().getValue()+"; ");
				}
				System.out.println();
			}
			System.out.println();
			System.out.println("--- Machines ---");
			for(int i = 0; i< nmchs; i++){	
				for (int j = 0; j< ntsks; j++) {
					System.out.print(" " + mx[i][j].getStart().getValue() + " to " + mx[i][j].getEnd().getValue()+"; ");
				}
				System.out.println();
			}
		//sv.printStatistics();
		}
		System.out.println("\n !!! No (more) solutions after " + sv.getTimeCount() + " secs");
	}
}	

/*	
filters:
• DEFAULT: Combines above filters as a black-box not idempotent
• DISJUNCTIVE_TASK_INTERVAL: energetic reasoning to filter disjunctive constraint Only propagated on variable subsets of size < 30 not idempotent not enough to ensure correctness (only an additional filtering)
• HEIGHTS:filters height variables only (sweep-based algorithm) idempotent (on the given set of variables only)
• NAIVETIME:time-table algorithm based on each point in time not idempotent
• NRJ: energetic reasoning to filter not idempotent not enough to ensure correctness (only an additional filtering)
• SWEEP: time-table algorithm based on a sweep line idempotent (on the given set of variables only)
• SWEEP_HEI_SORT: time-table algorithm based on a sweep line idempotent (on the given set of variables only)
• TIME: time-table algorithm based on each point in time not idempotent		
		
graphBased - parameter indicating how to filter: 
	TRUE: applies on subset of overlapping tasks
	FALSE: applies on all tasks
*/

